#!/bin/sh
set -xe

# only relevant on Debian
dpkg-vendor --is Debian || exit 0

# enable systemd-udevd
mount -o remount,rw /sys
systemctl unmask systemd-udevd.service
systemctl start systemd-udevd.service
